from django.core.management.base import BaseCommand
from defusedxml import ElementTree as ET

from django_gar.models import GARInstitution
from django_gar.gar import get_gar_institution_list


class Command(BaseCommand):
    help = "Refresh ENT IDs for all institutions from GAR"

    def handle(self, *args, **options):
        institutions = GARInstitution.objects.all()

        response = get_gar_institution_list()
        if response.status_code != 200:
            self.stderr.write(
                self.style.ERROR(
                    f"Failed to get institution list for GAR: {response.status_code}"
                )
            )
            return

        # Parse XML once
        root = ET.fromstring(response.content)
        namespace = {"ns": "http://www.atosworldline.com/listEtablissement/v1.0/"}
        gar_institutions = {
            etab.find("ns:uai", namespace).text: etab.find("ns:idENT", namespace).text
            for etab in root.findall("ns:etablissement", namespace)
        }

        updates = []
        for institution in institutions:
            if institution.uai in gar_institutions:
                institution.id_ent = gar_institutions[institution.uai]
                updates.append(institution)

        GARInstitution.objects.bulk_update(updates, ["id_ent"])



import pytest
from django.core.management import call_command
from django_gar.models import GARInstitution

pytestmark = pytest.mark.django_db


class TestRefreshGarCachesCommand:
    @pytest.mark.usefixtures(
        "mock_get_allocations_response",
        "mock_gar_request_response",
        "mock_gar_institution_list_response",
    )
    def test_refresh_caches_for_all_institutions(self, user, capsys):
        # GIVEN
        # Create another institution
        second_user = user.__class__.objects.create(username="test2")
        second_institution = GARInstitution.objects.create(
            user=second_user,
            uai="0123456A",
            institution_name="Test Institution 2",
            subscription_id="test_id_2",
        )

        # WHEN
        call_command("refresh_gar_caches")
        captured = capsys.readouterr()

        # THEN
        assert "Refreshing caches for 2 institution(s)..." in captured.out
        assert f"Processing {user.garinstitution.uai}..." in captured.out
        assert f"Processing {second_institution.uai}..." in captured.out
        assert "Done!" in captured.out

    @pytest.mark.usefixtures(
        "mock_get_allocations_response",
        "mock_gar_request_response",
        "mock_gar_institution_list_response",
        "mock_get_gar_subscription",
    )
    def test_refresh_caches_for_specific_uai(self, user, capsys):
        # GIVEN
        # Create another institution
        second_user = user.__class__.objects.create(username="test2")
        GARInstitution.objects.create(
            user=second_user,
            uai="0123456A",
            institution_name="Test Institution 2",
            subscription_id="test_id_2",
        )

        # WHEN
        call_command("refresh_gar_caches", uai=user.garinstitution.uai)
        captured = capsys.readouterr()

        # THEN
        assert "Refreshing caches for 1 institution(s)..." in captured.out
        assert f"Processing {user.garinstitution.uai}..." in captured.out
        assert "Done!" in captured.out

    def test_refresh_caches_with_invalid_uai(self, capsys):
        # WHEN
        call_command("refresh_gar_caches", uai="INVALID")
        captured = capsys.readouterr()

        # THEN
        assert 'No institution found with UAI "INVALID"' in captured.out
        assert "Done!" not in captured.out

    @pytest.mark.usefixtures(
        "mock_get_allocations_response",
        "mock_gar_request_response",
        "mock_gar_institution_list_response",
        "mock_get_gar_subscription",
    )
    def test_refresh_caches_updates_institution_data(self, user):
        # GIVEN
        institution = user.garinstitution
        assert institution.allocations_cache is None
        assert institution.subscription_cache is None

        # WHEN
        call_command("refresh_gar_caches", uai=institution.uai)

        # THEN
        institution.refresh_from_db()
        assert institution.allocations_cache["cumulAffectationEnseignant"] == "226"
        assert "abonnement" in institution.subscription_cache


class TestRefreshGarIdentsCommand:
    @pytest.mark.usefixtures(
        "mock_gar_institution_list_response", "mock_gar_request_response"
    )
    def test_refresh_idents_updates_institutions(self, user):
        # GIVEN
        institution = user.garinstitution
        institution.uai = "0941295X"
        institution.id_ent = None
        institution.save()

        # Create another institution
        second_user = user.__class__.objects.create(username="test2")
        second_institution = GARInstitution.objects.create(
            user=second_user,
            uai="0123456A",  # This one won't be found in GAR
            institution_name="Test Institution 2",
            subscription_id="test_id_2",
        )

        # WHEN
        call_command("refresh_gar_idents")

        # THEN
        institution.refresh_from_db()
        second_institution.refresh_from_db()
        assert institution.id_ent == "123456"
        assert second_institution.id_ent is None



import pytest

from django.conf import settings

from django_gar.middleware import GARMiddleware

pytestmark = pytest.mark.django_db

GAR_QUERY_STRING_TRIGGER = settings.GAR_QUERY_STRING_TRIGGER


class TestGARMiddleware:
    @staticmethod
    def test_init():
        """Testing the __init__ method"""
        # WHEN
        cas_middleware = GARMiddleware("dump_response")

        # THEN
        assert cas_middleware.get_response == "dump_response"

    def test_no_uai_no_ticket(self, request_builder):
        """Testing the middleware if not a connection to GAR"""
        # GIVEN
        cas_middleware = GARMiddleware(request_builder.get)

        # WHEN
        response = cas_middleware(request_builder.get())

        # THEN
        assert cas_middleware.get_response().path == "/"
        assert GAR_QUERY_STRING_TRIGGER not in response.GET
        assert "ticket" not in response.GET

    def test_when_uai_number(self, user, request_builder):
        """Testing the __call__ method with uai_number in url"""
        # GIVEN
        uai_number = "GAR"
        query_params = "?{}={}".format(GAR_QUERY_STRING_TRIGGER, uai_number)
        request = request_builder.get(query_params)
        cas_middleware = GARMiddleware(request)

        # WHEN
        response = cas_middleware(request)

        # THEN
        assert settings.GAR_BASE_URL.format(uai_number) in response.url

    def test_when_uai_number_is_gar(self, request_builder):
        """Testing the __call__ method with uai_number in url and is gar"""
        # GIVEN
        query_params = "?{}={}".format(GAR_QUERY_STRING_TRIGGER, "gar")
        request = request_builder.get(query_params)
        cas_middleware = GARMiddleware(request)

        # WHEN
        response = cas_middleware(request)

        # THEN
        assert settings.GAR_BASE_URL in response.url

    def test_when_cas_ticket_valid(self, mock_validate_valid_ticket, request_builder):
        """
        Testing the __call__ method with valid cas_ticket in url and user
        has access (is_active is True)
        """
        # GIVEN
        cas_ticket = "this-is-a-ticket"
        query_params = "?ticket={}".format(cas_ticket)
        request = request_builder.get(path=query_params)
        request.session["is_gar"] = True
        cas_middleware = GARMiddleware(request_builder.get)

        # WHEN
        response = cas_middleware(request)

        # THEN
        assert mock_validate_valid_ticket.call_count == 1
        assert response.url == settings.GAR_ACTIVE_USER_REDIRECT

    def test_when_cas_ticket_valid_and_grain(
        self, mock_validate_valid_ticket, request_builder
    ):
        """
        Redirect the user to a specific url using the 'grain' query string
        """
        # GIVEN
        cas_ticket = "this-is-a-ticket"
        grain_url = "https://www.dummy.com"
        query_params = "?ticket={}&grain={}".format(cas_ticket, grain_url)
        request = request_builder.get(path=query_params)
        request.session["is_gar"] = True
        cas_middleware = GARMiddleware(request_builder.get)

        # WHEN
        response = cas_middleware(request)

        # THEN
        assert mock_validate_valid_ticket.call_count == 1
        assert response.url == grain_url

    def test_when_cas_ticket_invalid(
        self, mock_validate_invalid_ticket, request_builder
    ):
        """
        Testing the __call__ method with invalid cas_ticket in url and user
        has access (is_active is True)
        """
        # GIVEN
        cas_ticket = "this-is-a-ticket"
        query_params = "/?ticket={}".format(cas_ticket)
        request = request_builder.get(query_params)
        request.session["is_gar"] = True
        cas_middleware = GARMiddleware(request_builder.get)

        # WHEN
        response = cas_middleware(request)

        # THEN
        assert response.status_code == 302
        assert mock_validate_invalid_ticket.call_count == 1
        assert response.url == settings.GAR_INACTIVE_USER_REDIRECT

    def test_validate_ticket(self, mock_verification_response, request_builder):
        # GIVEN
        request = request_builder.get()
        request.session["is_gar"] = True

        # WHEN
        mock_verification_response.get()
        cas_middleware = GARMiddleware(request_builder.get)
        uai_numbers = cas_middleware.validate_ticket(request, "dummy-ticket")

        # THEN
        assert "0561641E" in uai_numbers
        mock_verification_response.assert_called_once()

    def test_validate_ticket_parse_error(
        self, mock_verification_response_error, request_builder
    ):
        # GIVEN
        request = request_builder.get()
        request.session["is_gar"] = True

        # WHEN
        cas_middleware = GARMiddleware(request_builder.get)
        response = cas_middleware.validate_ticket(request, "dummy-ticket")

        # THEN
        assert not response



# Generated by Django 4.2.19 on 2025-02-13 13:08

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_gar", "0007_garinstitution_allocations_cache_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="garinstitution",
            name="subscription_cache",
            field=models.JSONField(blank=True, null=True, verbose_name="Abonnement"),
        ),
        migrations.AddField(
            model_name="garinstitution",
            name="subscription_cache_updated_at",
            field=models.DateTimeField(
                blank=True,
                null=True,
                verbose_name="Dernière mise à jour du cache abonnement",
            ),
        ),
    ]



# Generated by Django 2.2.28 on 2022-05-10 08:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_gar", "0003_auto_20201001_1054"),
    ]

    operations = [
        migrations.CreateModel(
            name="GARSession",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "ticket",
                    models.CharField(
                        max_length=255, unique=True, verbose_name="CAS ticket"
                    ),
                ),
                (
                    "session_key",
                    models.CharField(
                        max_length=255, unique=True, verbose_name="Django session key"
                    ),
                ),
            ],
        ),
    ]



from django import forms
from django.conf import settings
from django.contrib.auth import get_user_model
from django.forms import ModelForm
import unicodedata

from .models import GARInstitution

GAR_DISTRIBUTOR_ID = getattr(settings, "GAR_DISTRIBUTOR_ID", "")
GAR_RESOURCES_ID = getattr(settings, "GAR_RESOURCES_ID", "")
GAR_ORGANIZATION_NAME = getattr(settings, "GAR_ORGANIZATION_NAME", "")
GAR_SUBSCRIPTION_PREFIX = getattr(settings, "GAR_SUBSCRIPTION_PREFIX", "")
User = get_user_model()


class GARInstitutionForm(ModelForm):
    class Meta:
        model = GARInstitution
        fields = (
            "uai",
            "institution_name",
            "ends_at",
            "user",
            "project_code",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Disable ends_at field on update form
        instance = getattr(self, "instance", None)
        if instance.pk:
            self.fields["ends_at"].disabled = True

    def clean_uai(self):
        return self.cleaned_data.get("uai").upper().strip()

    def clean_institution_name(self):
        institution_name = self.cleaned_data.get("institution_name").upper()

        return "".join(
            c
            for c in unicodedata.normalize("NFD", institution_name)
            if unicodedata.category(c) != "Mn"
        )



import pytest

from django_gar.apps import DjangoGarConfig

pytestmark = pytest.mark.django_db


class TestDjangoGarConfig:
    @staticmethod
    def test_apps():
        assert "django_gar" in DjangoGarConfig.name



# Generated by Django 2.2.15 on 2020-08-21 14:05

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="GARInstitution",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uai",
                    models.CharField(
                        max_length=14,
                        unique=True,
                        verbose_name="Unité Administrative Immatriculée",
                    ),
                ),
                (
                    "institution_name",
                    models.CharField(
                        max_length=255, verbose_name="Nom de l'institution"
                    ),
                ),
                (
                    "ends_at",
                    models.DateField(
                        null=True, verbose_name="Date de fin d'abonnement"
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



from django.core.management.base import BaseCommand
from django_gar.models import GARInstitution


class Command(BaseCommand):
    help = "Refresh GAR subscription, allocations and ID ENT caches for institutions"

    def add_arguments(self, parser):
        parser.add_argument(
            "--uai",
            help="Refresh caches only for this UAI",
        )

    def handle(self, *args, **options):
        queryset = GARInstitution.objects.all()
        if options["uai"]:
            queryset = queryset.filter(uai=options["uai"])
            if not queryset.exists():
                self.stdout.write(
                    self.style.ERROR(
                        f'No institution found with UAI "{options["uai"]}"'
                    )
                )
                return

        count = queryset.count()
        self.stdout.write(f"Refreshing caches for {count} institution(s)...")

        for institution in queryset:
            self.stdout.write(f"Processing {institution.uai}...")

            institution.refresh_subscription_cache()
            institution.refresh_allocations_cache()

        self.stdout.write(self.style.SUCCESS("Done!"))



import csv

from django.urls import path
from django.conf import settings
from django.contrib import admin, messages
from django.http import HttpResponse
from django.shortcuts import redirect, reverse
from django.utils.html import format_html

from .gar import get_allocations
from .forms import GARInstitutionForm
from .models import GARInstitution

GAR_RESOURCES_ID = getattr(settings, "GAR_RESOURCES_ID", "")


@admin.register(GARInstitution)
class GARInstitutionAdmin(admin.ModelAdmin):
    raw_id_fields = ("user",)
    list_display = (
        "institution_name",
        "user",
        "uai",
        "ends_at",
        "student_allocation",
        "teacher_allocation",
    )
    list_select_related = ("user",)
    readonly_fields = (
        "id_ent",
        "subscription_id",
        "gar_subscription_response",
        "get_allocations",
    )
    ordering = ("institution_name",)
    search_fields = ("institution_name", "user__email", "uai", "project_code")
    list_filter = ["project_code"]
    form = GARInstitutionForm
    change_list_template = "admin/django_gar/change_list.html"

    @admin.display(description="Etat de l'abonnement dans le GAR")
    def gar_subscription_response(self, obj):
        if not obj.uai:
            return ""

        # Update cache if needed
        if not obj.subscription_cache:
            obj.refresh_subscription_cache()

        if not obj.subscription_cache:
            return (
                "L'abonnement n'existe pas dans le GAR. "
                "Vous pouvez le supprimer et en créer un nouveau."
            )

        # Group values by key in case there are multiple values
        grouped_values = {}
        for key, value in obj.subscription_cache.items():
            if key not in grouped_values:
                grouped_values[key] = []
            grouped_values[key].append(value)

        response = ""
        for key, values in grouped_values.items():
            # If there's only one value, display it normally
            if len(values) == 1:
                response += f"{key} : {values[0]}<br/>"
            # If there are multiple values, display them as a list
            else:
                response += f"{key} :<br/>"
                for value in values:
                    response += f"&nbsp;&nbsp;&nbsp;&nbsp;- {value}<br/>"

        return format_html(f"<code>{response}</code>")

    @admin.display(description="Etat des affectations")
    def get_allocations(self, obj):
        if not obj.uai or not obj.subscription_id:
            return ""

        # Update cache if needed
        if not obj.allocations_cache:
            obj.refresh_allocations_cache()

        if not obj.allocations_cache:
            return "L'établissement n'a pas encore affecté la ressource. Les informations fournies par le webservice font l'objet d'un traitement asynchrone et sont par conséquent actualisées quotidiennement. Il peut être constaté une latence dans la prise en compte de changements en cas d'affectations / récupérations de licences au sein d'une même journée."

        allocations = ""
        for key, value in obj.allocations_cache.items():
            allocations += f"{key} : {value}<br/>"

        return format_html(f"<code>{allocations}</code>")

    @admin.display(
        description="Affectations élève",
        ordering="allocations_cache__cumulAffectationEleve"
    )
    def student_allocation(self, obj):
        if not obj.allocations_cache:
            return "-"
        value = obj.allocations_cache.get("cumulAffectationEleve", "")
        try:
            return int(value) if value else 0
        except (ValueError, TypeError):
            return value or "-"

    @admin.display(
        description="Affectations enseignant",
        ordering="allocations_cache__cumulAffectationEnseignant"
    )
    def teacher_allocation(self, obj):
        if not obj.allocations_cache:
            return "-"
        value = obj.allocations_cache.get("cumulAffectationEnseignant", "")
        try:
            return int(value) if value else 0
        except (ValueError, TypeError):
            return value or "-"


    def get_urls(self):
        urlpatterns = super().get_urls()

        allocations_report_url = [
            path(
                "allocations-report/generate/",
                self.admin_site.admin_view(self.allocations_report),
                name="{app_label}_{model_name}_generate_allocations_report".format(
                    app_label=self.model._meta.app_label,
                    model_name=self.model._meta.model_name,
                ),
            )
        ]

        return allocations_report_url + urlpatterns

    def allocations_report(self, request):
        """Generate a CSV report of allocations for all institutions"""
        # Headers for the CSV
        headers = [
            "InstitutionName",
            "UAI",
            "idAbonnement",
            "codeProjetRessource",
            "idRessource",
            "cumulAffectationEleve",
            "cumulAffectationEnseignant",
            "cumulAffectationDocumentaliste",
            "cumulAffectationAutrePersonnel",
        ]

        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = (
            'attachment; filename="rapport_affectations.csv"'
        )
        writer = csv.writer(response)
        writer.writerow(headers)

        # Get all institutions
        institutions = self.model.objects.all()

        # Write data for each institution
        for institution in institutions:
            allocations = institution.allocations_cache or {}
            writer.writerow(
                [
                    institution.institution_name,
                    institution.uai,
                    institution.subscription_id,
                    institution.project_code,
                    settings.GAR_RESOURCES_ID,
                    allocations.get("cumulAffectationEleve", "0"),
                    allocations.get("cumulAffectationEnseignant", "0"),
                    allocations.get("cumulAffectationDocumentaliste", "0"),
                    allocations.get("cumulAffectationAutrePersonnel", "0"),
                ]
            )

        return response



import pytest

from django.contrib.admin.sites import AdminSite

from django_gar.admin import GARInstitutionAdmin
from django_gar.models import GARInstitution

pytestmark = pytest.mark.django_db


class TestGARInstitutionAdmin:
    def test_gar_subscription_response_no_uai(self, user):
        # GIVEN
        gar_institution = user.garinstitution
        gar_institution.uai = None
        gar_institution_admin = GARInstitutionAdmin(GARInstitution, AdminSite())

        # WHEN
        response = gar_institution_admin.gar_subscription_response(gar_institution)

        # THEN
        assert response == ""

    @pytest.mark.usefixtures("mock_get_gar_subscription_empty")
    def test_gar_subscription_response_no_subscription(self, user):
        # GIVEN
        gar_institution = user.garinstitution
        gar_institution_admin = GARInstitutionAdmin(GARInstitution, AdminSite())

        # WHEN
        response = gar_institution_admin.gar_subscription_response(gar_institution)

        # THEN
        assert response == (
            "L'abonnement n'existe pas dans le GAR. "
            "Vous pouvez le supprimer et en créer un nouveau."
        )

    @pytest.mark.usefixtures("mock_get_gar_subscription")
    def test_gar_subscription_response_with_data(self, user):
        # GIVEN
        gar_institution = user.garinstitution
        gar_institution.subscription_id = "briefme_1630592238"
        gar_institution.uai = "0561622J"
        gar_institution_admin = GARInstitutionAdmin(GARInstitution, AdminSite())

        # WHEN
        response = gar_institution_admin.gar_subscription_response(gar_institution)

        # THEN
        assert "idabonnement : briefme_1630592238<br/>" in response
        assert "typeaffectation : INDIV<br/>" in response

    def test_get_allocations_no_uai(self, user):
        # GIVEN
        gar_institution = user.garinstitution
        gar_institution.uai = None
        gar_institution_admin = GARInstitutionAdmin(GARInstitution, AdminSite())

        # WHEN
        response = gar_institution_admin.get_allocations(gar_institution)

        # THEN
        assert response == ""

    @pytest.mark.usefixtures("mock_get_allocations_empty_response")
    def test_get_allocations_with_empty_allocation(self, user):
        # GIVEN
        gar_institution = user.garinstitution
        gar_institution_admin = GARInstitutionAdmin(GARInstitution, AdminSite())

        # WHEN
        response = gar_institution_admin.get_allocations(gar_institution)

        # THEN
        assert "L'établissement n'a pas encore affecté la ressource" in response

    @pytest.mark.usefixtures(
        "mock_get_allocations_response",
        "mock_gar_request_response",
        "mock_gar_institution_list_response",
    )
    def test_get_allocations_with_csv_data(self, user):
        # GIVEN
        gar_institution = user.garinstitution
        gar_institution_admin = GARInstitutionAdmin(GARInstitution, AdminSite())

        # WHEN
        response = gar_institution_admin.get_allocations(gar_institution)

        # THEN
        assert "idAbonnement : briefme_0941295X_1709542737.2902117<br/>" in response
        assert "cumulAffectationEnseignant : 226<br/>" in response



# Generated by Django 3.2.16 on 2023-01-03 13:54

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("django_gar", "0004_garsession"),
    ]

    operations = [
        migrations.AddField(
            model_name="garinstitution",
            name="project_code",
            field=models.CharField(
                blank=True,
                max_length=50,
                null=True,
                verbose_name="Code de projet ressources",
            ),
        ),
    ]



import logging
import urllib.parse

from bs4 import BeautifulSoup

from django.contrib.sessions.backends.db import SessionStore
from django.http import HttpResponse
from django.template.response import TemplateResponse
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from .models import GARSession

logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name="dispatch")
class LogoutView(View):
    def get(self, request, *args, **kwargs):
        return TemplateResponse(request, "django_gar/logout.html", {})

    def post(self, request, *args, **kwargs):
        """x-www-form-urlencoded body to decode from GARrequest logout"""
        body = request.body.decode("utf-8")

        logger.info(body)

        data = urllib.parse.parse_qs(body)["logoutRequest"][0]

        soup = BeautifulSoup(data, "xml")
        session_index = soup.find("samlp:SessionIndex")

        if session_index:
            try:
                gar_session = GARSession.objects.get(ticket=session_index.text)
                store = SessionStore(session_key=gar_session.session_key)
                logger.info(f"request GAR logout {gar_session.session_key}")

                store["gar_logout"] = True
                store.save()
            except GARSession.DoesNotExist:
                logger.info(
                    "cannot log out from GAR session as it does not exist anymore"
                )

        return HttpResponse(status=204)



from setuptools import setup, find_packages

from django_gar import __version__

from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="django-gar",
    version=__version__,
    description="Handle login and ticket validation for french GAR",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/briefmnews/django-gar",
    author="Brief.me",
    author_email="tech@brief.me",
    license="GNU GPL v3",
    packages=find_packages(exclude=["tests"]),
    python_requires=">=3.9",
    install_requires=[
        "Django>=4.2",
        "python-cas>=1.6.0",
        "lxml>=4.9.4",
        "defusedxml>=0.7.1",
        "requests>=2.29.0",
    ],
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
    ],
    include_package_data=True,
    zip_safe=False,
)



import pytest
import requests
from urllib.parse import urlencode

from django_gar.exceptions import DjangoGARException
from django_gar.gar import (
    get_gar_subscription,
    get_allocations,
    GAR_ALLOCATIONS_URL,
    get_gar_subscription_end_date,
    get_gar_institution_list,
    delete_gar_subscription,
    GAR_BASE_SUBSCRIPTION_URL,
)

pytestmark = pytest.mark.django_db()


class TestGetGarSubscription:
    def test_with_subscription_in_response(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.uai = "0561622J"
        user.garinstitution.subscription_id = "briefeco_1630592291"
        mocker.patch("django_gar.signals.handlers.handle_gar_subscription")
        with open(
            "tests/fixtures/get_gar_subscription_response.xml", "r"
        ) as xml_response:
            mock_request = mocker.patch.object(
                requests,
                "request",
                return_value=response_from_gar(200, xml_response.read()),
            )

        # WHEN
        response = get_gar_subscription(
            user.garinstitution.uai, user.garinstitution.subscription_id
        )

        # THEN
        assert mock_request.call_count == 1
        assert response

    def test_with_wrong_subscription_id(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.uai = "0561622J"
        mocker.patch("django_gar.signals.handlers.handle_gar_subscription")
        with open(
            "tests/fixtures/get_gar_subscription_response.xml", "r"
        ) as xml_response:
            mock_request = mocker.patch.object(
                requests,
                "request",
                return_value=response_from_gar(200, xml_response.read()),
            )

        # WHEN
        response = get_gar_subscription(
            user.garinstitution.uai, user.garinstitution.subscription_id
        )

        # THEN
        assert mock_request.call_count == 1
        assert not response

    def test_with_status_code_not_200(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.subscription_id = "briefeco_1630592291"
        mocker.patch("django_gar.signals.handlers.handle_gar_subscription")
        with open(
            "tests/fixtures/get_gar_subscription_response.xml", "r"
        ) as xml_response:
            mock_request = mocker.patch.object(
                requests,
                "request",
                return_value=response_from_gar(404, xml_response.read()),
            )

        # WHEN / THEN
        with pytest.raises(DjangoGARException):
            get_gar_subscription(
                user.garinstitution.uai, user.garinstitution.subscription_id
            )

        assert mock_request.call_count == 1

    def test_with_empty_response(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.subscription_id = "briefeco_1630592291"
        mocker.patch("django_gar.signals.handlers.handle_gar_subscription")
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(200, ""),
        )

        # WHEN
        response = get_gar_subscription(
            user.garinstitution.uai, user.garinstitution.subscription_id
        )

        # THEN
        assert mock_request.call_count == 1
        assert not response


class TestGetGarSubscriptionEndDate:
    def test_with_valid_subscription(self, user, mocker, response_from_gar):
        # GIVEN
        expected_date = "2024-12-31"
        xml_content = """<?xml version="1.0" encoding="UTF-8"?>
        <abonnements xmlns="http://www.atosworldline.com/wsabonnement/v1.0/">
            <abonnement>
                <idabonnement>briefeco_1630592291</idabonnement>
                <finvalidite>2024-12-31</finvalidite>
            </abonnement>
        </abonnements>"""
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(200, xml_content),
        )

        # WHEN
        end_date = get_gar_subscription_end_date(
            user.garinstitution.uai, "briefeco_1630592291"
        )

        # THEN
        assert mock_request.called_once()
        assert end_date == expected_date

    def test_with_no_subscription(self, user, mocker, response_from_gar):
        # GIVEN
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(200, "<abonnements></abonnements>"),
        )

        # WHEN
        end_date = get_gar_subscription_end_date(
            user.garinstitution.uai, "non_existent_id"
        )

        # THEN
        assert mock_request.called_once()
        assert end_date is None


class TestGetAllocations:
    def test_get_allocations_no_params(self):
        # GIVEN / WHEN
        with pytest.raises(DjangoGARException) as exc_info:
            get_allocations()

        # THEN
        assert exc_info.value.status_code == 400
        assert (
            exc_info.value.message
            == "At least one of subscription_id or project_code is mandatory"
        )

    def test_get_allocations_both_params(self):
        # GIVEN / WHEN
        with pytest.raises(DjangoGARException) as exc_info:
            get_allocations(subscription_id="123", project_code="ABC")

        # THEN
        assert exc_info.value.status_code == 400
        assert (
            exc_info.value.message
            == "Cannot set subscription_id and project_code at the same time"
        )

    def test_get_allocations_with_subscription_id(
        self, user, mock_get_allocations_request
    ):
        # WHEN
        response = get_allocations(subscription_id=user.garinstitution.subscription_id)

        # THEN
        mock_get_allocations_request.assert_called_once_with(
            "GET",
            f"{GAR_ALLOCATIONS_URL}?idAbonnement={user.garinstitution.subscription_id}",
            cert=("", ""),
        )
        assert response.status_code == 200

    def test_get_allocations_with_project_code(self, mock_get_allocations_request):
        # WHEN
        response = get_allocations(project_code="DUMMY")

        # THEN
        mock_get_allocations_request.assert_called_once_with(
            "GET", f"{GAR_ALLOCATIONS_URL}?codeProjetRessource=DUMMY", cert=("", "")
        )
        assert response.status_code == 200

    def test_with_subscription_id(self, mocker, response_from_gar):
        # GIVEN
        subscription_id = "test_sub_id"
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(200, "response"),
        )

        # WHEN
        response = get_allocations(subscription_id=subscription_id)

        # THEN
        assert mock_request.call_count == 1
        expected_url = (
            f"{GAR_ALLOCATIONS_URL}?{urlencode({'idAbonnement': subscription_id})}"
        )
        mock_request.assert_called_with(
            "GET",
            expected_url,
            cert=mocker.ANY,
        )

    def test_with_project_code(self, mocker, response_from_gar):
        # GIVEN
        project_code = "test_project"
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(200, "response"),
        )

        # WHEN
        response = get_allocations(project_code=project_code)

        # THEN
        assert mock_request.call_count == 1
        expected_url = (
            f"{GAR_ALLOCATIONS_URL}?{urlencode({'codeProjetRessource': project_code})}"
        )
        mock_request.assert_called_with(
            "GET",
            expected_url,
            cert=mocker.ANY,
        )

    def test_with_no_parameters(self):
        # WHEN/THEN
        with pytest.raises(DjangoGARException) as exc:
            get_allocations()
        assert "At least one of subscription_id or project_code is mandatory" in str(
            exc.value
        )

    def test_with_both_parameters(self):
        # WHEN/THEN
        with pytest.raises(DjangoGARException) as exc:
            get_allocations(subscription_id="test", project_code="test")
        assert "Cannot set subscription_id and project_code at the same time" in str(
            exc.value
        )


class TestDeleteGarSubscription:
    def test_successful_deletion(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.subscription_id = "test_subscription"
        mock_request = mocker.patch.object(
            requests,
            "delete",
            return_value=response_from_gar(200, "Successfully deleted"),
        )

        # WHEN
        delete_gar_subscription(user.garinstitution.subscription_id)

        # THEN
        assert mock_request.call_count == 1
        mock_request.assert_called_with(
            f"{GAR_BASE_SUBSCRIPTION_URL}test_subscription",
            cert=mocker.ANY,
            headers=mocker.ANY,
            timeout=30,
        )

    def test_failed_deletion(self, user, mocker, response_from_gar):
        # GIVEN
        user.garinstitution.subscription_id = "test_subscription"
        mock_request = mocker.patch.object(
            requests,
            "delete",
            return_value=response_from_gar(404, "Not found"),
        )

        # WHEN
        delete_gar_subscription(user.garinstitution.subscription_id)

        # THEN
        assert mock_request.call_count == 1
        mock_request.assert_called_with(
            f"{GAR_BASE_SUBSCRIPTION_URL}test_subscription",
            cert=mocker.ANY,
            headers=mocker.ANY,
            timeout=30,
        )


class TestGetGarInstitutionList:
    def test_successful_list_retrieval(self, mocker, response_from_gar):
        # GIVEN
        mock_request = mocker.patch.object(
            requests,
            "request",
            return_value=response_from_gar(200, "Institution list"),
        )

        # WHEN
        response = get_gar_institution_list()

        # THEN
        assert mock_request.call_count == 1
        assert response.status_code == 200



import logging

from django.db import models
from django.contrib.auth import get_user_model
from django.utils import timezone

from .gar import get_allocations, get_gar_subscription


logger = logging.getLogger(__name__)
User = get_user_model()


class GARInstitution(models.Model):
    uai = models.CharField(
        "Unité Administrative Immatriculée", max_length=14, unique=True
    )
    institution_name = models.CharField("Nom de l'institution", max_length=255)
    id_ent = models.CharField("ID de l'ent", max_length=255, null=True)
    ends_at = models.DateField("Date de fin d'abonnement", null=True)
    user = models.OneToOneField(User, on_delete=models.CASCADE)
    subscription_id = models.CharField("id abonnement", max_length=255, unique=True)
    project_code = models.CharField(
        "Code de projet ressources", max_length=50, null=True, blank=True
    )
    allocations_cache = models.JSONField("Affectations", null=True, blank=True)
    allocations_cache_updated_at = models.DateTimeField(
        "Dernière mise à jour du cache", null=True, blank=True
    )
    subscription_cache = models.JSONField("Abonnement", null=True, blank=True)
    subscription_cache_updated_at = models.DateTimeField(
        "Dernière mise à jour du cache abonnement", null=True, blank=True
    )

    def __str__(self):
        return f"{self.institution_name} ({self.uai})"

    def refresh_allocations_cache(self):
        if not self.subscription_id:
            return

        response = get_allocations(subscription_id=self.subscription_id)

        self.allocations_cache_updated_at = timezone.now()

        if response.status_code == 200:
            # Parse CSV data into a dictionary
            csv_content = response.content.decode("utf-8")
            lines = csv_content.splitlines()

            if len(lines) > 1:  # If we have header and data
                headers = lines[0].split(";")
                values = lines[1].split(";")  # We take the first data line
                self.allocations_cache = dict(zip(headers, values))
            else:
                self.allocations_cache = None

            logger.info("Allocations cache updated successfully.")
        else:
            self.allocations_cache = None
            logger.error(
                f"Failed to refresh allocations cache. Status code: {response.status_code}, Response: {response.text}"
            )

        self.save(
            update_fields=["allocations_cache", "allocations_cache_updated_at"],
        )

    def refresh_subscription_cache(self):
        if not self.uai or not self.subscription_id:
            return

        subscription = get_gar_subscription(self.uai, self.subscription_id)

        self.subscription_cache_updated_at = timezone.now()

        if subscription:
            # Create a dictionary to store all values
            cache = {}
            for element in subscription.find_all():
                if element.text.strip():  # Only store non-empty text
                    if element.name in cache:
                        # If we already have this key and it's not a list, convert to list
                        if not isinstance(cache[element.name], list):
                            cache[element.name] = [cache[element.name]]
                        cache[element.name].append(element.text)
                    else:
                        cache[element.name] = element.text

            self.subscription_cache = cache
            logger.info("Subscription cache updated successfully.")
        else:
            self.subscription_cache = None
            logger.error(f"No subscription found in GAR for {self.uai}.")

        self.save(update_fields=["subscription_cache", "subscription_cache_updated_at"])


class GARSession(models.Model):
    """Store GAR active session. This will help us to delete user sessions when the user log out from the GAR"""

    ticket = models.CharField("CAS ticket", max_length=255, unique=True)
    session_key = models.CharField("Django session key", max_length=255, unique=True)

    def __str__(self):
        return f"ticket: {self.ticket} - session_key: {self.session_key}"
